/*
 * FreeRTOS Kernel V10.0.1
 * Copyright (C) 2017 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * http://www.FreeRTOS.org
 * http://aws.amazon.com/freertos
 *
 * 1 tab == 4 spaces!
 */


// -- includes -- //

#include "FreeRTOS.h"
#include "task.h"

#include "portmacro.h"
#include "FPGAConfig.h"
#include "riscvdef.h"
#include "hal.h"

// -- macros --//

#ifdef __riscv64
# define STORE    sd
# define LOAD     ld
# define REGBYTES 8
#else
# define STORE    sw
# define LOAD     lw
# define REGBYTES 4
#endif

// -- global variables & function declarations -- //

uint32_t l_CriticalNesting = 0xaaaaaaaa;
uint32_t l_SchedulerStarted = 0;
void vPortSetupTimer(void);
void prvTaskExitError(void);

// -- function definitions -- //

void vPortSetupTimer(void) {
	
	TimerSetPeriod(g_Timer, configCPU_CLOCK_HZ / configTICK_RATE_HZ);
	
	// Timer will not be started here - this happens when the soft interrupt is raised.
	
	/* Set bit 0 of register MEIMASK (Orca Core) to enable timer interrupts.
	 * set_csr(reg, bit) stringifies the register and inserts into asm() directive.
	 * Since MEIMASK doesn't exist in standart riscv the compilation fails.
	 * Therefore the set_csr_ns macro is used to first convert MEIMASK to 0x7C0 an then insert into asm().
	 */
	
	// Soft Interrupt Timer init
	g_SITimer->periodl = 1;
	g_SITimer->periodh = 0;
	
	set_csr_ns(MEIMASK, IRQ_TIMER_BIT);
	
}

void prvTaskExitError(void) {
	
	/*
	A function that implements a task must not exit or attempt to return to
	its caller as there is nothing to return to.  If a task wants to exit it
	should instead call vTaskDelete( NULL ).

	Program execution is halted and debug information is put on UART.
	*/

	portDISABLE_INTERRUPTS();
	DebugWrite("Error: Task Exit\n");
	PrintStack(40);
	while (1) {
		// TODO: Add error handling code.
	}
}

void vPortClearInterruptMask(int mask) {
	__asm volatile("csrw mie, %0"::"r"(mask));
}

int vPortSetInterruptMask(void) {
	int ret;
	__asm volatile("csrr %0,mie":"=r"(ret));
	__asm volatile("csrc mie,%0"::"i"(7));
	return ret;
}

void vPortYield(void) {
	halRaiseSoftInterrupt();
}

void vPortEnterCritical(void) {
	portDISABLE_INTERRUPTS();
	l_CriticalNesting++;
}

void vPortExitCritical(void) {
	l_CriticalNesting--;
	if (l_CriticalNesting == 0) {
		portENABLE_INTERRUPTS();
	}
}

StackType_t *pxPortInitialiseStack(StackType_t* pxTopOfStack, TaskFunction_t pxCode, void* pvParameters) {

	//Simulate Stack frame as it would be created by interrupt. Compare to crt.S or SysTick_Handler.
	
	pxTopOfStack -= 32; // move by 32 * 4 Bytes (size of StackType_t)
	pxTopOfStack[0] = (portSTACK_TYPE) prvTaskExitError; // x1, ra - return address
	//pxTopOfStack[3] = (portSTACK_TYPE) tp; // Thread pointers are currently not supported.
	pxTopOfStack[9] = (portSTACK_TYPE) pvParameters; // x10, a0 - first parameter
	pxTopOfStack[31] = (portSTACK_TYPE) pxCode; // mepc - return adress after interrupt
	return pxTopOfStack;
}

BaseType_t xPortStartScheduler(void) {
	set_csr_ns(MEIMASK, IRQ_SOFT_BIT);
	vPortSetupTimer();
	l_SchedulerStarted = 1;
	l_CriticalNesting = 0;
	halEnableInterrupts();
	halRaiseSoftInterrupt();
	// Scheduler should never return to this line
	return pdFALSE;
}

void vPortEndScheduler(void) {/* Do not implement. */}

void SysTick_Handler(uintptr_t stackPointer) {
	
	if (l_SchedulerStarted == 1){
		l_SchedulerStarted = 2;
	} else {
		asm volatile("lw t0, pxCurrentTCB");
		asm volatile("sw a0, 0(t0)");
	}
	
	if (xTaskIncrementTick() != pdFALSE) {
		vTaskSwitchContext();
	}
		
	__asm volatile("lw	 sp, pxCurrentTCB");
	__asm volatile("lw	 sp, 0x0(sp)");
	__asm volatile("lw	 t0, 31 * 4(sp)");
	__asm volatile("csrw	mepc, t0");
	__asm volatile("lw	 x1, 0(sp)");
	//__asm volatile("lw	 x2, 1 * 4(sp)"); //  Stack Pointer - do not restore here
	//__asm volatile("lw	 x3, 2 * 4(sp)"); // Global Pointer - do not restore
	__asm volatile("lw   x4, 3 * 4(sp)");
	__asm volatile("lw   x5, 4 * 4(sp)");
	__asm volatile("lw   x6, 5 * 4(sp)");
	__asm volatile("lw   x7, 6 * 4(sp)");
	__asm volatile("lw   x8, 7 * 4(sp)");
	__asm volatile("lw   x9, 8 * 4(sp)");
	__asm volatile("lw   x10, 9 * 4(sp)");
	__asm volatile("lw   x11, 10 * 4(sp)");
	__asm volatile("lw   x12, 11 * 4(sp)");
	__asm volatile("lw   x13, 12 * 4(sp)");
	__asm volatile("lw   x14, 13 * 4(sp)");
	__asm volatile("lw   x15, 14 * 4(sp)");
	__asm volatile("lw   x16, 15 * 4(sp)");
	__asm volatile("lw   x17, 16 * 4(sp)");
	__asm volatile("lw   x18, 17 * 4(sp)");
	__asm volatile("lw   x19, 18 * 4(sp)");
	__asm volatile("lw   x20, 19 * 4(sp)");
	__asm volatile("lw   x21, 20 * 4(sp)");
	__asm volatile("lw   x22, 21 * 4(sp)");
	__asm volatile("lw   x23, 22 * 4(sp)");
	__asm volatile("lw   x24, 23 * 4(sp)");
	__asm volatile("lw   x25, 24 * 4(sp)");
	__asm volatile("lw   x26, 25 * 4(sp)");
	__asm volatile("lw   x27, 26 * 4(sp)");
	__asm volatile("lw   x28, 27 * 4(sp)");
	__asm volatile("lw   x29, 28 * 4(sp)");
	__asm volatile("lw   x30, 29 * 4(sp)");
	__asm volatile("lw   x31, 30 * 4(sp)");
	__asm volatile("addi	sp, sp, 4 * 32");	
	asm("csrs mstatus, 8");
	__asm volatile("mret");
	
}

