#include "FreeRTOS.h"
#include "task.h"
#include "queue.h"
#include "timers.h"
#include "FPGAConfig.h"
#include "hal.h"

/* A block time of zero simply means "don't block". */
#define mainDONT_BLOCK						( 0UL )

// Tasks
void taskGPIOCounter(void* param);
void taskLEDBlinking(void* param);

// IRQ Handlers
void UARTIRQHandler(void) {
	char c = UARTGet(g_UART);
	UARTPut(g_UART, c);
	g_UART->status = 0;
}

/*
 * FreeRTOS hook for when malloc fails, enable in FreeRTOSConfig.
 */
void vApplicationMallocFailedHook( void );

/*
 * FreeRTOS hook for when FreeRtos is idling, enable in FreeRTOSConfig.
 */
void vApplicationIdleHook( void );

/*
 * FreeRTOS hook for when a stack overflow occurs, enable in FreeRTOSConfig.
 */
void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName );

/*-----------------------------------------------------------*/

int main(void) {
	
	// Enable receive interrupt on UART
	g_UART->control = UART_RRDY;

	// Initialize Interrupt Control
	halSetInterruptHandler(UARTIRQHandler, IRQ_UART);
	halEnableIRQ(IRQ_UART_BIT);
	
	#if (HEAP_CONFIG == HEAP_SDRAM)
	// Initialize HEAP - 8 MiB (0x800000 Bytes) starting from address 0x0
	HeapRegion_t SDRAM[] = {
		{ ( uint8_t * ) 0x00000000, 0x800000 },
		{ NULL, 0 }
	};
	vPortDefineHeapRegions(SDRAM);
	#endif
	
	// RTOS Tasks	
	xTaskCreate(taskGPIOCounter, "GPIOCounter", 256, NULL, uartPRIMARY_PRIORITY, NULL);
	xTaskCreate(taskLEDBlinking, "LEDBlinking", 256, NULL, uartPRIMARY_PRIORITY, NULL);
	
	/* Start the kernel.  From here on, only tasks and interrupts will run. */
	vTaskStartScheduler();
	
	/* Exit FreeRTOS */
	return 0;
}

/*-----------------------------------------------------------*/

void vApplicationMallocFailedHook( void )
{
	/* vApplicationMallocFailedHook() will only be called if
	configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
	function that will get called if a call to pvPortMalloc() fails.
	pvPortMalloc() is called internally by the kernel whenever a task, queue,
	timer or semaphore is created.  It is also called by various parts of the
	demo application.  If heap_1.c or heap_2.c are used, then the size of the
	heap available to pvPortMalloc() is defined by configTOTAL_HEAP_SIZE in
	FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
	to query the size of free heap space that remains (although it does not
	provide information on how the remaining heap might be fragmented). */
	taskDISABLE_INTERRUPTS();
	DebugWrite("Malloc Failure\n");
	while (true) {
		// TODO: Add failure recovery code
	}
}
/*-----------------------------------------------------------*/

void vApplicationIdleHook( void )
{
	/* vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
	to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
	task.  It is essential that code added to this hook function never attempts
	to block in any way (for example, call xQueueReceive() with a block time
	specified, or call vTaskDelay()).  If the application makes use of the
	vTaskDelete() API function (as this demo application does) then it is also
	important that vApplicationIdleHook() is permitted to return to its calling
	function, because it is the responsibility of the idle task to clean up
	memory allocated by the kernel to any task that has since been deleted. */
}
/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook( TaskHandle_t pxTask, char *pcTaskName )
{
	( void ) pcTaskName;
	( void ) pxTask;

	/* Run time stack overflow checking is performed if
	configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook
	function is called if a stack overflow is detected. */
	taskDISABLE_INTERRUPTS();
	DebugWrite("StackOverflow\n");
	while (true) {
		// TODO: Add failure recovery code
	}
}
/*-----------------------------------------------------------*/

void taskGPIOCounter(void* param) {
	g_GPIO->direction = 0xFF;
	while (true) {
		delay(100 * MS_TO_CLOCK);
		g_GPIO->gpio++;
	}
}

void taskLEDBlinking(void* param){

	uint32_t timePrev = get_time();
	uint32_t timeNow;

	while (true) {
		do {
			timeNow = get_time();
		} while ((timeNow - timePrev) < (1000 * MS_TO_CLOCK));

		timePrev = timeNow;
		g_LED->green ^= 1;	
	}
}